/*
** $Id: liolib.c,v 2.73.1.4 2010/05/14 15:33:51 roberto Exp $
** Standard I/O (and system) library
** See Copyright Notice in lua.h
*/


#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define liolib_c
#define LUA_LIB

#include "lua.h"

#include "lauxlib.h"
#include "lualib.h"

FILE* default_fopen(lua_State *L, const char * filename, const char * mode) {return fopen(filename, mode);}
int default_fclose(lua_State *L, FILE * stream) {return fclose(stream);}

FILE* (*mounter_fopen)(lua_State *L, const char * filename, const char * mode) = &default_fopen;
int (*mounter_fclose)(lua_State *L, FILE * stream) = &default_fclose;

#define fopen(f, m) mounter_fopen(L, f, m)
#define fclose(f) mounter_fclose(L, f)


#define IO_INPUT	1
#define IO_OUTPUT	2


static const char *const fnames[] = {"input", "output"};


static int pushresult (lua_State *L, int i, const char *filename) {
  int en = errno;  /* calls to Lua API may change this value */
  if (i) {
    lua_pushboolean(L, 1);
    return 1;
  }
  else {
    lua_pushnil(L);
    if (filename)
      lua_pushfstring(L, "%s: %s", filename, strerror(en));
    else
      lua_pushfstring(L, "%s", strerror(en));
    lua_pushinteger(L, en);
    return 3;
  }
}


static void fileerror (lua_State *L, int arg, const char *filename) {
  lua_pushfstring(L, "%s: %s", filename, strerror(errno));
  luaL_argerror(L, arg, lua_tostring(L, -1));
}


#define tofilep(L)	((FILE **)luaL_checkudata(L, 1, LUA_FILEHANDLE))

static char stdfile_id = 'S';

static int isstdfile(lua_State *L, int idx) {
    lua_getmetatable(L, idx);
    if (!lua_istable(L, -1)) {
        lua_pop(L, 1);
        return 0;
    }
    lua_getfield(L, -1, "__fileid");
    int res = lua_isuserdata(L, -1) && lua_touserdata(L, -1) == &stdfile_id;
    lua_pop(L, 2);
    return res;
}

static int io_type (lua_State *L) {
  void *ud;
  luaL_checkany(L, 1);
  ud = lua_touserdata(L, 1);
  lua_getfield(L, LUA_REGISTRYINDEX, LUA_FILEHANDLE);
  if (isstdfile(L, 1))
    lua_pushliteral(L, "file");
  else if (ud == NULL || !lua_getmetatable(L, 1) || !lua_rawequal(L, -2, -1))
    lua_pushnil(L);  /* not a file */
  else if (*((FILE **)ud) == NULL)
    lua_pushliteral(L, "closed file");
  else
    lua_pushliteral(L, "file");
  return 1;
}


static FILE *tofile (lua_State *L) {
  FILE **f = tofilep(L);
  if (*f == NULL)
    luaL_error(L, "attempt to use a closed file");
  return *f;
}



/*
** When creating file handles, always creates a `closed' file handle
** before opening the actual file; so, if there is a memory error, the
** file is not left opened.
*/
static FILE **newfile (lua_State *L) {
  FILE **pf = (FILE **)lua_newuserdata(L, sizeof(FILE *));
  *pf = NULL;  /* file handle is currently `closed' */
  luaL_getmetatable(L, LUA_FILEHANDLE);
  lua_setmetatable(L, -2);
  return pf;
}


/*
** function to (not) close the standard files stdin, stdout, and stderr
*/
static int io_noclose (lua_State *L) {
  lua_pushnil(L);
  lua_pushliteral(L, "attempt to close standard stream");
  return 2;
}


/*
** function to close regular files
*/
static int io_fclose (lua_State *L) {
  FILE **p = tofilep(L);
  int ok = (fclose(*p) == 0);
  *p = NULL;
  return pushresult(L, ok, NULL);
}


static int aux_close (lua_State *L) {
  lua_getfenv(L, 1);
  lua_getfield(L, -1, "__close");
  return (lua_tocfunction(L, -1))(L);
}


static int io_close (lua_State *L) {
  if (lua_isnone(L, 1))
    lua_rawgeti(L, LUA_ENVIRONINDEX, IO_OUTPUT);
  tofile(L);  /* make sure argument is a file */
  return aux_close(L);
}


static int io_gc (lua_State *L) {
  FILE *f = *tofilep(L);
  /* ignore closed files */
  if (f != NULL)
    aux_close(L);
  return 0;
}


static int io_tostring (lua_State *L) {
  FILE *f = *tofilep(L);
  if (f == NULL)
    lua_pushliteral(L, "file (closed)");
  else
    lua_pushfstring(L, "file (%p)", f);
  return 1;
}


static int io_open (lua_State *L) {
  if (!lua_isstring(L, 1)) luaL_error(L, "bad argument #1 (expected string, got %s)", lua_typename(L, lua_type(L, 1)));
  if (!lua_isstring(L, 2) && !lua_isnoneornil(L, 2)) luaL_error(L, "bad argument #2 (expected string, got %s)", lua_typename(L, lua_type(L, 2)));
  const char *filename = luaL_checkstring(L, 1);
  const char *mode = luaL_optstring(L, 2, "r");
  FILE **pf = newfile(L);
  *pf = fopen(filename, mode);
  return (*pf == NULL) ? pushresult(L, 0, filename) : 1;
}


static FILE *getiofile (lua_State *L, int findex) {
  FILE *f;
  lua_rawgeti(L, LUA_ENVIRONINDEX, findex);
  if (!lua_isuserdata(L, -1)) return NULL;
  f = *(FILE **)lua_touserdata(L, -1);
  if (f == NULL)
    luaL_error(L, "standard %s file is closed", fnames[findex - 1]);
  return f;
}


static int g_iofile (lua_State *L, int f, const char *mode) {
  if (!lua_isnoneornil(L, 1)) {
    const char *filename = lua_tostring(L, 1);
    if (filename) {
      FILE **pf = newfile(L);
      *pf = fopen(filename, mode);
      if (*pf == NULL)
        fileerror(L, 1, filename);
    }
    else {
      if (!lua_istable(L, 1)) {
        tofile(L);  /* check that it's a valid file handle */
        lua_pushvalue(L, 1);
      }
    }
    lua_rawseti(L, LUA_ENVIRONINDEX, f);
  }
  /* return current value */
  lua_rawgeti(L, LUA_ENVIRONINDEX, f);
  return 1;
}


static int io_input (lua_State *L) {
  return g_iofile(L, IO_INPUT, "r");
}


static int io_output (lua_State *L) {
  return g_iofile(L, IO_OUTPUT, "w");
}


static int io_readline (lua_State *L);


static void aux_lines (lua_State *L, int idx, int toclose) {
  int n, i;
  for (n = 1; lua_isstring(L, idx + n) || lua_isnumber(L, idx + n); n++);
  lua_pushvalue(L, idx);
  lua_pushboolean(L, toclose);  /* close/not close file when finished */
  if (n == 1) {
    lua_pushinteger(L, n++);
    lua_pushstring(L, "*l");
  } else {
    lua_pushinteger(L, n - 1);
    for (i = 1; i < n; i++) lua_pushvalue(L, idx + i);
  }
  lua_pushcclosure(L, io_readline, n + 2);
}


static int f_lines (lua_State *L) {
  if (isstdfile(L, 1)) {
    lua_getglobal(L, "read");
  } else {
    tofile(L);  /* check that it's a valid file handle */
    aux_lines(L, 1, 0);
  }
  return 1;
}

#include <assert.h>

static int io_lines (lua_State *L) {
  if (lua_isnoneornil(L, 1)) {  /* no arguments? */
    /* will iterate over default input */
    lua_rawgeti(L, LUA_ENVIRONINDEX, IO_INPUT);
    if (lua_isnil(L, 1)) lua_replace(L, 1);
    return f_lines(L);
  }
  else {
    int i;
    const char *filename = luaL_checkstring(L, 1);
    FILE **pf = newfile(L);
    *pf = fopen(filename, "r");
    if (*pf == NULL)
      fileerror(L, 1, filename);
    lua_replace(L, 1);
    aux_lines(L, 1, 1);
    return 1;
  }
}


/*
** {======================================================
** READ
** =======================================================
*/


static int read_number (lua_State *L, FILE *f) {
  lua_Number d;
  if (fscanf(f, LUA_NUMBER_SCAN, &d) == 1) {
    lua_pushnumber(L, d);
    return 1;
  }
  else {
    lua_pushnil(L);  /* "result" to be removed */
    return 0;  /* read fails */
  }
}


static int test_eof (lua_State *L, FILE *f) {
  int c = getc(f);
  ungetc(c, f);
  lua_pushlstring(L, NULL, 0);
  return (c != EOF);
}


static int read_line (lua_State *L, FILE *f, int keepnl) {
  luaL_Buffer b;
  luaL_buffinit(L, &b);
  for (;;) {
    size_t l;
    char *p = luaL_prepbuffer(&b);
    if (fgets(p, LUAL_BUFFERSIZE, f) == NULL) {  /* eof? */
      luaL_pushresult(&b);  /* close buffer */
      return (lua_objlen(L, -1) > 0);  /* check whether read something */
    }
    l = strlen(p);
    if (l == 0 || p[l-1] != '\n')
      luaL_addsize(&b, l);
    else {
      luaL_addsize(&b, keepnl ? l : l - 1);  /* do not include `eol' unless asked for */
      luaL_pushresult(&b);  /* close buffer */
      return 1;  /* read at least an `eol' */
    }
  }
}


static int read_chars (lua_State *L, FILE *f, size_t n) {
  size_t rlen;  /* how much to read */
  size_t nr;  /* number of chars actually read */
  luaL_Buffer b;
  luaL_buffinit(L, &b);
  rlen = LUAL_BUFFERSIZE;  /* try to read that much each time */
  do {
    char *p = luaL_prepbuffer(&b);
    if (rlen > n) rlen = n;  /* cannot read more than asked */
    nr = fread(p, sizeof(char), rlen, f);
    luaL_addsize(&b, nr);
    n -= nr;  /* still have to read `n' chars */
  } while (n > 0 && nr == rlen);  /* until end of count or eof */
  luaL_pushresult(&b);  /* close buffer */
  return (n == 0 || lua_objlen(L, -1) > 0);
}


static int g_read (lua_State *L, FILE *f, int first) {
  int nargs = lua_gettop(L) - 1;
  int success;
  int n;
  if (f == NULL) {
    lua_rawgeti(L, LUA_ENVIRONINDEX, IO_INPUT);
    lua_getfield(L, -1, "read");
    lua_insert(L, -nargs-3);
    lua_pop(L, 2);
    lua_call(L, nargs, 1);
    return 1;
  }
  clearerr(f);
  if (nargs == 0) {  /* no arguments? */
    success = read_line(L, f, 0);
    n = first+1;  /* to return 1 result */
  }
  else {  /* ensure stack space for all results and for auxlib's buffer */
    luaL_checkstack(L, nargs+LUA_MINSTACK, "too many arguments");
    success = 1;
    for (n = first; nargs-- && success; n++) {
      if (lua_type(L, n) == LUA_TNUMBER) {
        size_t l = (size_t)lua_tointeger(L, n);
        success = (l == 0) ? test_eof(L, f) : read_chars(L, f, l);
      }
      else {
        const char *p = lua_tostring(L, n);
        luaL_argcheck(L, p, n, "invalid option");
        if (*p == '*') p++;
        switch (*p) {
          case 'n':  /* number */
            success = read_number(L, f);
            break;
          case 'l':  /* line */
            success = read_line(L, f, 0);
            break;
          case 'L':  /* line (with newline) */
            success = read_line(L, f, 1);
            break;
          case 'a':  /* file */
            read_chars(L, f, ~((size_t)0));  /* read MAX_SIZE_T chars */
            success = 1; /* always success */
            break;
          default:
            return luaL_argerror(L, n, "invalid format");
        }
      }
    }
  }
  if (ferror(f))
    return pushresult(L, 0, NULL);
  if (!success) {
    lua_pop(L, 1);  /* remove last result */
    lua_pushnil(L);  /* push nil instead */
  }
  return n - first;
}


static int io_read (lua_State *L) {
  return g_read(L, getiofile(L, IO_INPUT), 1);
}


static int f_read (lua_State *L) {
  return g_read(L, tofile(L), 2);
}


static int io_readline (lua_State *L) {
  FILE *f = *(FILE **)lua_touserdata(L, lua_upvalueindex(1));
  int success, i;
  const char * p;
  if (f == NULL)  /* file is already closed? */
    return luaL_error(L, "file is already closed");
  for (i = 0; i < lua_tointeger(L, lua_upvalueindex(3)); i++) {
    if (lua_isnumber(L, lua_upvalueindex(i + 4))) {
      size_t l = (size_t)lua_tointeger(L, lua_upvalueindex(i + 4));
      success = (l == 0) ? test_eof(L, f) : read_chars(L, f, l);
    } else {
      p = lua_tostring(L, lua_upvalueindex(i + 4));
      if (p == NULL) return luaL_error(L, "invalid option #%d to 'lines'", i + 1);
      if (*p == '*') p++;
      switch (*p) {
        case 'n':  /* number */
          success = read_number(L, f);
          break;
        case 'l':  /* line */
          success = read_line(L, f, 0);
          break;
        case 'L':  /* line (with newline) */
          success = read_line(L, f, 1);
          break;
        case 'a':  /* file */
          read_chars(L, f, ~((size_t)0));  /* read MAX_SIZE_T chars */
          success = 1; /* always success */
          break;
        default:
          return luaL_error(L, "invalid format #%d to 'lines'", i + 1);
      }
    }
    if (ferror(f))
      return luaL_error(L, "%s", strerror(errno));
    if (!success) {  /* EOF */
      if (lua_toboolean(L, lua_upvalueindex(2))) {  /* generator created file? */
        lua_settop(L, 0);
        lua_pushvalue(L, lua_upvalueindex(1));
        aux_close(L);  /* close it */
      }
      return 0;
    }
  }
  return i;
}

/* }====================================================== */


static int g_write (lua_State *L, FILE *f, int arg) {
  int nargs = lua_gettop(L) - 1;
  int status = 1;
  if (f == NULL) {
    lua_rawgeti(L, LUA_ENVIRONINDEX, IO_OUTPUT);
    lua_getfield(L, -1, "write");
    lua_insert(L, -nargs-3);
    lua_pop(L, 2);
    lua_call(L, nargs, 1);
    return 1;
  }
  for (; nargs--; arg++) {
    if (lua_type(L, arg) == LUA_TNUMBER) {
      /* optimization: could be done exactly as for strings */
      status = status &&
          fprintf(f, LUA_NUMBER_FMT, lua_tonumber(L, arg)) > 0;
    }
    else {
      size_t l;
      const char *s = luaL_checklstring(L, arg, &l);
      status = status && (fwrite(s, sizeof(char), l, f) == l);
    }
  }
  return status;
}


static int io_write (lua_State *L) {
  int i, en = errno; /* calls to Lua API may change this value */
  i = g_write(L, getiofile(L, IO_OUTPUT), 1);
  if (i) {
    lua_rawgeti(L, LUA_ENVIRONINDEX, IO_OUTPUT);
    return 1;
  } else {
    lua_pushnil(L);
    lua_pushfstring(L, "%s", strerror(en));
    lua_pushinteger(L, en);
    return 3;
  }
}


static int f_write (lua_State *L) {
  int i, en = errno;  /* calls to Lua API may change this value */
  i = g_write(L, tofile(L), 2);
  if (i) {
    lua_pushvalue(L, 1);
    return 1;
  } else {
    lua_pushnil(L);
    lua_pushfstring(L, "%s", strerror(en));
    lua_pushinteger(L, en);
    return 3;
  }
}


static int f_seek (lua_State *L) {
  static const int mode[] = {SEEK_SET, SEEK_CUR, SEEK_END};
  static const char *const modenames[] = {"set", "cur", "end", NULL};
  FILE *f = tofile(L);
  int op = luaL_checkoption(L, 2, "cur", modenames);
  long offset = luaL_optlong(L, 3, 0);
  op = fseek(f, offset, mode[op]);
  if (op)
    return pushresult(L, 0, NULL);  /* error */
  else {
    lua_pushinteger(L, ftell(f));
    return 1;
  }
}


static int f_setvbuf (lua_State *L) {
  static const int mode[] = {_IONBF, _IOFBF, _IOLBF};
  static const char *const modenames[] = {"no", "full", "line", NULL};
  FILE *f = tofile(L);
  int op = luaL_checkoption(L, 2, NULL, modenames);
  lua_Integer sz = luaL_optinteger(L, 3, LUAL_BUFFERSIZE);
  int res = setvbuf(f, NULL, mode[op], sz);
  return pushresult(L, res == 0, NULL);
}



static int io_flush (lua_State *L) {
  return pushresult(L, fflush(getiofile(L, IO_OUTPUT)) == 0, NULL);
}


static int f_flush (lua_State *L) {
  return pushresult(L, fflush(tofile(L)) == 0, NULL);
}


static const luaL_Reg iolib[] = {
  {"close", io_close},
  {"flush", io_flush},
  {"input", io_input},
  {"lines", io_lines},
  {"open", io_open},
  {"output", io_output},
  {"read", io_read},
  {"type", io_type},
  {"write", io_write},
  {NULL, NULL}
};


static const luaL_Reg flib[] = {
  {"close", io_close},
  {"flush", f_flush},
  {"lines", f_lines},
  {"read", f_read},
  {"seek", f_seek},
  {"setvbuf", f_setvbuf},
  {"write", f_write},
  {"__gc", io_gc},
  {"__tostring", io_tostring},
  {NULL, NULL}
};


static void createmeta (lua_State *L) {
  luaL_newmetatable(L, LUA_FILEHANDLE);  /* create metatable for file handles */
  lua_pushvalue(L, -1);  /* push metatable */
  lua_setfield(L, -2, "__index");  /* metatable.__index = metatable */
  lua_pushstring(L, "FILE*");  /* push FILE* */
  lua_setfield(L, -2, "__name");  /* metatable.__name = "FILE*" (as per Cobalt spec) */
  luaL_register(L, NULL, flib);  /* file methods */
}


static void newfenv (lua_State *L, lua_CFunction cls) {
  lua_createtable(L, 0, 1);
  lua_pushcfunction(L, cls);
  lua_setfield(L, -2, "__close");
}

int stdin_read_count = 0, stdout_read_count = 0, stderr_read_count = 0;

static int stdfile_seek(lua_State *L) {
  const char * whence;
  int val, newval;
  int * read_count;
  if (lua_istable(L, 1)) lua_remove(L, 1);
  read_count = (int*)lua_touserdata(L, lua_upvalueindex(1));
  if (!lua_isnoneornil(L, 1)) {
    whence = luaL_checkstring(L, 1);
    if (strcmp(whence, "cur") && strcmp(whence, "set") && strcmp(whence, "end")) 
      luaL_error(L, "bad argument #1 to 'seek' (invalid option '%s')", whence);
    if (!lua_isnoneornil(L, 2)) {
      val = luaL_checkint(L, 2);
      if (strcmp(whence, "cur") == 0) newval = *read_count + val;
      else newval = val;
      if (newval < 0) {
        lua_pushnil(L);
        lua_pushstring(L, "Invalid argument");
        lua_pushinteger(L, 22);
        return 3;
      }
      *read_count = newval;
    } else if (strcmp(whence, "set") == 0 || strcmp(whence, "end") == 0) *read_count = 0;
  }
  lua_pushinteger(L, *read_count);
  return 1;
}

static int stdin_read(lua_State *L) {
  if (lua_istable(L, 1)) lua_remove(L, 1);
  luaL_optstring(L, 1, "*l");
  if (lua_isstring(L, 1) && strcmp(lua_tostring(L, 1), "*l")) luaL_error(L, "bad argument #1 to 'read' (invalid format)");
  lua_getglobal(L, "read");
  lua_call(L, 0, 1);
  stdin_read_count += lua_isstring(L, -1) ? lua_strlen(L, -1) : 0;
  return 1;
}

static int stdin_lines(lua_State *L) {lua_pushcfunction(L, stdin_read); return 1;}

static int aux_badfd(lua_State *L) {
  lua_pushnil(L);
  lua_pushstring(L, "file is not writable");
  return 2;
}

static int stdout_write(lua_State *L) {
  if (lua_istable(L, 1)) lua_remove(L, 1);
  lua_getglobal(L, "write");
  lua_insert(L, 1);
  lua_call(L, lua_gettop(L)-1, 0);
  return 0;
}

static int stderr_write(lua_State *L) {
  if (lua_istable(L, 1)) lua_remove(L, 1);
  lua_getglobal(L, "term");
  lua_getfield(L, -1, "getTextColor");
  lua_call(L, 0, 1);
  int color = lua_tointeger(L, -1);
  lua_pop(L, 1);
  lua_getfield(L, -1, "setTextColor");
  lua_pushinteger(L, 16384);
  lua_call(L, 1, 0);
  lua_getglobal(L, "write");
  lua_pushvalue(L, -3);
  lua_call(L, 1, 0);
  lua_getfield(L, -1, "setTextColor");
  lua_pushinteger(L, color);
  lua_call(L, 1, 0);
  lua_pop(L, 1);
  return 0;
}

static int aux_true(lua_State *L) {lua_pushboolean(L, 1); return 1;}

LUALIB_API int luaopen_io (lua_State *L) {
  createmeta(L);
  /* create (private) environment (with fields IO_INPUT, IO_OUTPUT, __close) */
  newfenv(L, io_fclose);
  lua_replace(L, LUA_ENVIRONINDEX);
  /* open library */
  luaL_register(L, LUA_IOLIBNAME, iolib);
  /* create (and set) default files */
  // ComputerCraft stdin
  lua_newtable(L);
  lua_newtable(L);
  lua_pushstring(L, "FILE*");
  lua_setfield(L, -2, "__name");
  lua_pushlightuserdata(L, &stdfile_id);
  lua_setfield(L, -2, "__fileid");
  lua_setmetatable(L, -2);
  lua_pushcfunction(L, io_noclose);
  lua_setfield(L, -2, "close");
  lua_pushcfunction(L, aux_true);
  lua_setfield(L, -2, "flush");
  lua_pushcfunction(L, stdin_lines);
  lua_setfield(L, -2, "lines");
  lua_pushcfunction(L, stdin_read);
  lua_setfield(L, -2, "read");
  lua_pushlightuserdata(L, &stdin_read_count);
  lua_pushcclosure(L, stdfile_seek, 1);
  lua_setfield(L, -2, "seek");
  lua_pushcfunction(L, aux_true);
  lua_setfield(L, -2, "setvbuf");
  lua_pushcfunction(L, aux_badfd);
  lua_setfield(L, -2, "write");
  lua_pushvalue(L, -1);
  lua_setfield(L, -3, "stdin");
  lua_rawseti(L, LUA_ENVIRONINDEX, IO_INPUT);
  // ComputerCraft stdout
  lua_newtable(L);
  lua_newtable(L);
  lua_pushstring(L, "FILE*");
  lua_setfield(L, -2, "__name");
  lua_pushlightuserdata(L, &stdfile_id);
  lua_setfield(L, -2, "__fileid");
  lua_setmetatable(L, -2);
  lua_pushcfunction(L, io_noclose);
  lua_setfield(L, -2, "close");
  lua_pushcfunction(L, aux_true);
  lua_setfield(L, -2, "flush");
  lua_pushcfunction(L, aux_badfd);
  lua_setfield(L, -2, "lines");
  lua_pushcfunction(L, aux_badfd);
  lua_setfield(L, -2, "read");
  lua_pushlightuserdata(L, &stdout_read_count);
  lua_pushcclosure(L, stdfile_seek, 1);
  lua_setfield(L, -2, "seek");
  lua_pushcfunction(L, aux_true);
  lua_setfield(L, -2, "setvbuf");
  lua_pushcfunction(L, stdout_write);
  lua_setfield(L, -2, "write");
  lua_pushvalue(L, -1);
  lua_setfield(L, -3, "stdout");
  lua_rawseti(L, LUA_ENVIRONINDEX, IO_OUTPUT);
  // ComputerCraft stderr
  lua_newtable(L);
  lua_newtable(L);
  lua_pushstring(L, "FILE*");
  lua_setfield(L, -2, "__name");
  lua_pushlightuserdata(L, &stdfile_id);
  lua_setfield(L, -2, "__fileid");
  lua_setmetatable(L, -2);
  lua_pushcfunction(L, io_noclose);
  lua_setfield(L, -2, "close");
  lua_pushcfunction(L, aux_true);
  lua_setfield(L, -2, "flush");
  lua_pushcfunction(L, aux_badfd);
  lua_setfield(L, -2, "lines");
  lua_pushcfunction(L, aux_badfd);
  lua_setfield(L, -2, "read");
  lua_pushlightuserdata(L, &stderr_read_count);
  lua_pushcclosure(L, stdfile_seek, 1);
  lua_setfield(L, -2, "seek");
  lua_pushcfunction(L, aux_true);
  lua_setfield(L, -2, "setvbuf");
  lua_pushcfunction(L, stderr_write);
  lua_setfield(L, -2, "write");
  lua_setfield(L, -2, "stderr");
  return 1;
}

LUALIB_API void lualib_io_ccpc_functions(FILE* (*open)(lua_State *L, const char * filename, const char * mode), int (*close)(lua_State *L, FILE * stream)) {
  mounter_fopen = open;
  mounter_fclose = close;
}

